import numpy as np
import pandas as pd
import itertools
import random
import os
import json
import pathlib
from models import *
from metrics import *
from train_utils import *
from train import *

curr_dir_path = str(pathlib.Path().absolute())
save_dir_path = '/logs_2d/' ##'/logs/'  ##''/logs/'  ##'/logs/'  ##
cwd = os.getcwd() 
FILE_NAME = 'CNN_hyp_tuning'

gpu_num = 0
cuda_available = torch.cuda.is_available()
if cuda_available:
    torch.cuda.set_device(gpu_num)

manual_seed = 1
if manual_seed is None:
    manual_seed = random.randint(1, 100000)
    
torch.manual_seed(manual_seed)
np.random.seed(manual_seed)
random.seed(manual_seed)

print("GPU num", gpu_num)
print("save at", save_dir_path)

def helper_func(obj):
    best_metrics = obj.training_pipeline()
    return best_metrics

if __name__ == '__main__':

    with open(curr_dir_path + "/config.json", encoding='utf-8', errors='ignore') as json_data:
        config = json.load(json_data, strict=False)

    hyp_batch_size = [32, 64, 128]
    hyp_no_conv = [1, 2] 
    hyp_no_filters = [16, 25, 32]
    hyp_filter_size = [(1, 5) (1,6), (1, 7)] 
    hyp_filter_size2 = [5, 6, 7] 
    hyp_stride = [(1,1)]
    hyp_pool_size = [(2, 2), (3, 3)]
    hyp_pool_stride = [(2, 2), (3, 3)] 
    hyp_dense_output = [64, 128]
    hyp_lr = [5e-5, 1e-4, 1e-3] 
    hyp_dropout = [0.0, 0.2, 0.5] 
    
    
    list_hyperparam = [hyp_batch_size, hyp_no_conv, hyp_no_filters, hyp_filter_size, hyp_stride, hyp_pool_size, hyp_pool_stride,
                       hyp_dense_output, hyp_lr, hyp_dropout, hyp_filter_size2]
    
    grid = list(itertools.product(*list_hyperparam))
    models_no = len(grid)
    print(f"The total number of models is {models_no}")
    random.shuffle(grid)
    hyp_file_write_path = cwd + save_dir_path + FILE_NAME 
    """
    hyp_cfg = pd.DataFrame(columns = ['Batch Size','Conv no','Filter no', 'Conv Filter size', 'Conv Filter stride', 
                                      'Pooling size', 'Pooling stride', 'Fully connected layer output', 'LR', 'Dropout', 'Val Loss', 
                                      'Val Accuracy', 'Val F1', 'Val Prec', 'Val Recall', 'Train Loss', 'Train Accuracy', 
                                      'Train F1', 'Train Prec', 'Train Recall'])
    
    
    hyp_cfg.to_csv(hyp_file_write_path, mode='a', header=True)
    """
    for i in range(0,models_no):
        
        hyp = grid[i]
        batch_size = hyp[0]; no_conv= hyp[1]; no_filters = hyp[2]; filter_size = hyp[3]; stride = hyp[4]; pool_size = hyp[5];
        pool_stride = hyp[6]; dense_output = hyp[7]; lr = hyp[8]; dropout = hyp[9]; filter_size2 = hyp[10]
        
        pool_stride = pool_size
        
        #Edit the config file   ##does not edit it 
        config['DATA']['BATCH_SIZE'] = batch_size
        config['MODEL']['Conv no'] = no_conv
        config['MODEL']['Filter no'] = no_filters
        config['MODEL']['Filter size'] = filter_size
        config['MODEL']['Filter size2'] = filter_size2
        config['MODEL']['Conv stride'] = stride
        config['MODEL']['Pool size'] = pool_size
        config['MODEL']['Pool stride'] = pool_stride
        config['MODEL']['FC layer output'] = dense_output
        config['MODEL']['dropout'] = dropout
        config['OPTIMIZER']['lr'] = lr
        
        # Train model        
        print('------------------------------')
        print(f'Start Training Model {i}')
        print('------------------------------')
        
        print(batch_size,
        no_conv,
        no_filters,
        filter_size,
        filter_size2,
        stride,
        pool_size,
        pool_stride,
        dense_output,
        lr,
        dropout)
            
        obj = Training(config,  save_dir_path, model_no=i)
        obj.training_pipeline()

        #Write hyper-parameters and results to csv file
        print('Writing hyper-parameters and results file "{}"'.format(hyp_file_write_path))
        metrics_train = obj.metrics['train']
        metrics_val = obj.metrics['val']
        data = {'Model': i, 'Batch Size': batch_size,'Conv no': no_conv,'Filter no': no_filters, 'Conv Filter size1': filter_size[0], 
                      'Conv Filter size1': filter_size[1], 'Conv Filter stride1': stride[0], 'Conv Filter stride2': stride[1],
                      'Pooling size': pool_size, 'Pooling stride':pool_stride, 
                     'Fully connected layer output': dense_output, 'LR':lr, 'Dropout':dropout, 'Val Loss': metrics_val['loss'], 
                     'Val Accuracy': metrics_val['acc'], 'Val F1': metrics_val['f1'], 'Val Prec': metrics_val['prec'], 
                     'Val Recall': metrics_val['recall'], 'Train Loss': metrics_train['loss'], 'Train Accuracy': metrics_train['acc'],
                     'Train F1': metrics_train['f1'], 'Train Prec': metrics_train['prec'], 'Train Recall': metrics_train['recall']}
        
       # output_folder = pathlib.Path(hyp_file_write_path)
       # output_folder.mkdir(parents=True, exist_ok=True)
       # data_json = json.dumps(data, indent=4, sort_keys=True)
       # (output_folder / f"results{self.model_no}.json").write_text(data_json)
            
        pd.DataFrame(data, index=[0]).to_csv(hyp_file_write_path, mode='a', header=False)
