import os
import numpy as np
import argparse
import pandas as pd
import time
import math
import json
import pathlib
import random

import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt
from torch.utils.data import DataLoader
from sklearn.model_selection import StratifiedShuffleSplit
from numpy import loadtxt, savetxt
from sklearn import metrics, model_selection
import seaborn as sns
from load_data import load_dataset

from vae_gumbel_softmax_old import VAE_gumbel, latent_repr, evaluate, VAE_gumbel_hp_tuning
from cnn.train import Training 
from cnn.train_utils import *
from cnn.models import *
from cnn.metrics import *



parser = argparse.ArgumentParser(description='PyTorch VAE GS')

parser.add_argument('--incl_rec', type=bool, default=False, metavar='incl_rec',
                    help='include reconstruction error')
parser.add_argument('-d', '--dataset', type=str, default='mnist', choices=['mnist'],
                    metavar='DATASET',
                    help='Dataset choice.')

parser.add_argument('-nc', '--no_cuda', action='store_true', default=False,
                    help='disables CUDA training')

parser.add_argument('--manual_seed', type=int, help='manual seed, if not given resorts to random seed.')

parser.add_argument('-li', '--log_interval', type=int, default=10, metavar='LOG_INTERVAL',
                    help='how many batches to wait before logging training status')

parser.add_argument('-od', '--out_dir', type=str, default='logs/', metavar='OUT_DIR',
                    help='output directory for model snapshots etc.')

fp = parser.add_mutually_exclusive_group(required=False)
fp.add_argument('-te', '--testing', action='store_true', dest='testing',
                help='evaluate on test set after training')
fp.add_argument('-va', '--validation', action='store_false', dest='testing',
                help='only evaluate on validation set')
parser.set_defaults(testing=False)

# optimization settings
parser.add_argument('-e', '--epochs', type=int, default=1, metavar='EPOCHS',
                    help='number of epochs to train (default: 1000)')
parser.add_argument('-es', '--early_stopping_epochs', type=int, default=100, metavar='EARLY_STOPPING',
                    help='number of early stopping epochs')

parser.add_argument('-bs', '--batch_size', type=int, default=1, metavar='BATCH_SIZE',
                    help='input batch size for training (default: 100)')
parser.add_argument('-lr', '--learning_rate', type=float, default=0.00001, metavar='LEARNING_RATE',
                    help='learning rate')

parser.add_argument('-a', '--anneal', type=str, default="kl", choices= ["std", "off", "kl"], help="beta annealing scheme")
parser.add_argument('--max_beta', type=float, default=1., metavar='MB',
                    help='max beta for warm-up')
parser.add_argument('--min_beta', type=float, default=0.0, metavar='MB',
                    help='min beta for warm-up')
parser.add_argument('-f', '--flow', type=str, default='planar', choices=['planar', 'NICE', 'NICE_MLP', 'syl_orthogonal', 'real' ])
parser.add_argument('-nf', '--num_flows', type=int, default=0,
                    metavar='NUM_FLOWS', help='Number of flow layers, ignored in absence of flows')
parser.add_argument('-nov', '--num_orthonormal_vec', type=int, default=8, metavar='NUM_ORTHONORMAL_VEC',
                    help='For orthogonal flow: number of orthogonal vectors (M)')
parser.add_argument('--syl_ortho_invertible', type=bool, default=False, metavar='ORTHO_FLOW_INVERT',
                    help='select if we need the sylvester orthogonal flow to be invertible')
parser.add_argument('--z_size', type=int, default=40, metavar='ZSIZE',
                    help='how many stochastic hidden units')
parser.add_argument('--flat_seq_size', type=int, default=320, metavar='FSSIZE',
                    help='size of flattened sequence')
parser.add_argument('--encoder_dim', type=int, default=40, metavar='ESIZE',
                    help='output feature dim of encoder')
parser.add_argument('--decoder_dim', type=int, default=320, metavar='DSIZE',
                    help='output feature dim of decoder') 

parser.add_argument('--cnn_enc', type=bool, default=True, metavar='cnn_enc',
                    help='encoder cnn')

parser.add_argument('--cnn_dec', type=bool, default=True, metavar='cnn_dec',
                    help='decoder cnn')
parser.add_argument('--test_model', type=bool, default=True, metavar='test_model',
                    help='test model')
# gpu/cpu
parser.add_argument('--gpu_num', type=int, default=0, metavar='GPU', help='choose GPU to run on.')

parser.add_argument('--temp', type=float, default=1.0, metavar='S',
                    help='tau(temperature) (default: 1.0)')


args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()

if args.manual_seed is None:
    args.manual_seed = random.randint(1, 100000)
random.seed(args.manual_seed)
torch.manual_seed(args.manual_seed)
np.random.seed(args.manual_seed)

if args.cuda:
    torch.cuda.set_device(args.gpu_num)

curr_dir_path = str(pathlib.Path().absolute())
save_dir_path = '/cnn/' + '/logs/'
gpu_num = 0 
cuda_available = torch.cuda.is_available()
if cuda_available:
    torch.cuda.set_device(gpu_num)

plt.rcParams.update(plt.rcParamsDefault)
#%matplotlib inline

plt.style.use('seaborn')
    
def run(args):
    mode = 'latent'
    test_dataset = 0
    
    args.length = 60
    
    length = args.length 
    width = 4*length #seq of 80 nucl one hot enc
    cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/starting_position/' 
      #  pos_seq_encoded_total_repeats = loadtxt(cwd + 'chr22.fa_positive_no_up_down_120.csv', delimiter=',')
    pos_seq_encoded_chr19 = loadtxt(cwd + 'chr21.fa_test_positive_80_spikes_stpos20.csv', delimiter=',')
        
    print(pos_seq_encoded_chr19.shape)
    cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/disease_repeats/'  
    disease = loadtxt(cwd + 'dataset_Disease_flanking.csv', delimiter=',') 
  #  disease = loadtxt(cwd + 'dataset_Marshfield_flanking.csv', delimiter=',') 
    end = disease.shape[1]
    disease = disease[:,0:end]
    

    
  #  seq = disease #= np.expand_dims(pos_seq_encoded_chr19[400:720], 0)

    pos_seq_encoded_chr19 = pos_seq_encoded_chr19
    pos_seq_encoded_chr19 = pos_seq_encoded_chr19.astype(np.float32)
    pos_seq_encoded_chr19 = torch.from_numpy(pos_seq_encoded_chr19)
    if test_dataset == 0:
        disease = pos_seq_encoded_chr19
        
    size = disease.shape[0]
    zer = torch.zeros(size, width)
    disease = np.concatenate((disease, zer), axis = 1)
    print("shape", disease.shape)
    disease = disease.astype(np.float32)
    disease = torch.from_numpy(disease)
    whole_seq = torch.reshape(disease, (1, -1))
    print(whole_seq.shape)
    whole_seq = whole_seq.squeeze(0)
    
    num = 2
    data = 0*whole_seq
    print(data.shape)
  
    #load VAE model

    dim = 30
    
    if args.incl_rec == True:
        dim += 1
        

  #  model = VAE_gumbel_hp_tuning(0.7, 6, 3, 16, 16, (1,7), (1,5), (1,1), (4,3), (1,3), length, 'sigmoid', 1)
    #json_dir = "model0.pt" ###model 0 sliding  - vae
    


    model = VAE_gumbel_hp_tuning(0.7, 10, 3, 12, 12, 5, 1, 2, 2, length, 'sigmoid', 0.2)
    json_dir = "model0_vaegs_before10.pt" 
    model.load_state_dict(torch.load(json_dir, map_location=torch.device('cpu')))   
    print(model)
    if mode == 'latent':
        #load MLP model for the latent space 
        with open(curr_dir_path + "/cnn" + "/config.json", encoding='utf-8', errors='ignore') as json_data:
            config = json.load(json_data, strict=False)   
        
        json_dir = curr_dir_path + "/model0_mlp_vaegs_before10.pt" ##"/model_mlp_sliding.pt"  #paired-model0 
      #  json_dir = curr_dir_path + "/cnn" + "/logs/" + "model0_mlp_epoch75.pt" #3curr_dir_path + save_dir_path + "model0.pt" 
        device =  torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        model_cnn = MLP(input_dim=dim, features=128)
        model_cnn.load_state_dict(torch.load(json_dir, map_location=torch.device('cpu')))
        print(model_cnn)   
        obj = Training(config, [], [], [], [], [], [], 4 , save_dir_path)
        obj.model = model_cnn
    
    spikes = np.zeros(int(data.shape[0]/4))
    binarized = np.zeros(int(data.shape[0]/4))
    spikes_mean = np.zeros(int(data.shape[0]/4))
    binarized_mean = np.zeros(int(data.shape[0]/4))
    
    zer = torch.zeros(width)
    whole_seq =  torch.cat([whole_seq, zer]) ### needed to be able to slide over the last window

    print(whole_seq.shape)
    
    if mode == 'latent':
        for i in range(0, data.shape[0] - width, 4):
            seq = whole_seq[i:i+width]   ##should 
            seq = seq.unsqueeze(0)
         #   print(i)

            seq = seq.cpu().detach().numpy() ##TODO not the best way        
            data2, args = load_dataset(args, seq)        
            latent2 = latent_repr(data2, model, args)        

          #  args.batch_size = 1
            if args.incl_rec == True:
             #   print("Including reconstruction error as an extra feature")
                loss2 = evaluate(data2, model, args)
                loss2 = np.expand_dims(loss2, 1)
                latent2 = np.concatenate((latent2, loss2), axis=1)

            #use the trained model on the latent space to make prediction
            latent2 = latent2.astype(np.float32)
            latent2 = torch.from_numpy(latent2) ##TODO not the best way 
            
            mask = np.isin(seq, 2)
            test = seq[mask]
            if (len(test) >0):
                spikes[int(i/4)] = 0
            else:
            
                output = obj.model.forward(latent2) 
                spikes[int(i/4)] = output.item()  
                
            if spikes[int(i/4)]>0.6:
                binarized[int(i/4)] = 1 
            else:
                binarized[int(i/4)] = 0
    
    elif mode == 'reconstruction_error':
        for i in range(0, data.shape[0] - width, 4):
            seq = whole_seq[i:i+width]   
            seq = seq.unsqueeze(0)

            seq = seq.cpu().detach().numpy()        
            data2, args = load_dataset(args, seq)        
            loss2 = evaluate(data2, model, args)
            loss2 = np.expand_dims(loss2, 1)
          
            mask = np.isin(seq, 2)
            test = seq[mask]
            if (len(test) >0):
                spikes[int(i/4)] = -100
            else:
                spikes[int(i/4)] = -loss2 

            if spikes[int(i/4)]>0.5:
                binarized[int(i/4)] = 1 
            else:
                binarized[int(i/4)] = 0
                
            if spikes[int(i/4)]>0.5:
                binarized[int(i/4)] = 1 
            else:
                binarized[int(i/4)] = 0
                
    print("spikes shape", spikes.shape)   
  
    spikes_avg = np.mean(spikes, axis=0)
    binarized_avg = np.mean(binarized, axis=0)
    for i in range(0, spikes.shape[0] - 5):
        spikes_mean[i] = np.mean(spikes[i:i+5])
        if spikes_mean[i]>0.5:
            binarized_mean[i] = 1 
        else:
            binarized_mean[i] = 0
            
    if test_dataset == 0: 
        length = 80
        if args.length == 60:
            spikes = spikes.reshape(-1, 3*length+60)       
            binarized = binarized.reshape(-1, 3*length+60)
            spikes_mean = spikes_mean.reshape(-1, 3*length+60)        
            binarized_mean = binarized_mean.reshape(-1, 3*length+60)
        else:
            spikes = spikes.reshape(-1, 4*length)       
            binarized = binarized.reshape(-1, 4*length)
            spikes_mean = spikes_mean.reshape(-1, 4*length)        
            binarized_mean = binarized_mean.reshape(-1, 4*length)
            
        length = args.length
        spikes = spikes[:,0:3*length]
        spikes_mean = spikes_mean[:,0:3*length]
        binarized = binarized[:,0:3*length]
        binarized_mean = binarized_mean[:,0:3*length]
        
        spikes_avg = np.mean(spikes, axis=0)
        binarized_avg = np.mean(binarized, axis=0)
        
        plt.plot(range(-40,140), binarized_avg)
        print(binarized_avg)
        print(spikes_avg)
      #  plt.plot(range(binarized.shape[0]), binarized_mean)
        plt.xlabel('Position', fontsize = 11) # x-axis label with fontsize 15
        plt.ylabel('Average model output', fontsize = 11)
        plt.yticks(rotation=0)
        
        plt.savefig('average_before_10_pos_gsvae.png', bbox_inches='tight') 
        plt.savefig('average_before_10_pos_gsvae.eps', bbox_inches='tight') 
        plt.show()
        
        
    else:
        end = int((end+width)/4)
        spikes = spikes.reshape(size, end)
        spikes_mean = spikes_mean.reshape( size, end)
        binarized = binarized.reshape(size, end)
        binarized_mean = binarized_mean.reshape(size, end) 
        
        spikes = spikes[:,0:end-1*length]
        spikes_mean = spikes_mean[:,0:end-1*length]
        binarized = binarized[:,0:end-1*length]
        binarized_mean = binarized_mean[:,0:end-1*length]
    
    
        spikes_avg = np.mean(spikes, axis=0)
        binarized_avg = np.mean(binarized, axis=0)
        print(spikes.shape)
        print(spikes_avg.shape)
        plt.plot(range(spikes_avg.shape[0]), spikes_avg)
        plt.plot(range(binarized_avg.shape[0]), binarized_avg)

        plt.show()
        """
        plt.plot(range(int((whole_seq.shape[0] - width)/4)), spikes)
        plt.plot(range(int((whole_seq.shape[0] - width)/4)), binarized)
        plt.show()
        spikes = np.expand_dims(spikes, 0)
        binarized = np.expand_dims(binarized, 0)
        """
    ### heatmap ###
    print(spikes_mean.shape)
 
    df = pd.DataFrame(spikes_mean, columns=range(-40,140))
    ax = sns.heatmap(df, vmin=0, vmax=1,  xticklabels=10)
    plt.xlabel('Position', fontsize = 11) # x-axis label with fontsize 15
    plt.ylabel('Sample num.', fontsize = 11)
    plt.yticks(rotation=0)
    
    ax.axes.yaxis.set_visible(False)
    fig = ax.get_figure()
    fig.savefig('heatmap_before__10_pos_gsvae.png', bbox_inches='tight') 
    fig.savefig('heatmap_before_10_pos_gsvae.eps', bbox_inches='tight') 
    plt.show()
    #plt.savefig(
    plt.show()
    print("position 0", spikes_avg[0])
    print("position 100", spikes_avg[100])
    print("position 120", spikes_avg[120])

  ###save date and plot spikes with actual repeats
if __name__ == "__main__":

    run(args)